@@ -22,11 +22,11 @@
 Information about the Frescobaldi application.
 """
 
-from __future__ import unicode_literals
 
 # these variables are also used by the distutils setup
 name = "frescobaldi"
-version = "2.20.0"
+version = "3.1"
+extension_api = "0.9.0"
 description = "LilyPond Music Editor"
 long_description = \
     "Frescobaldi is an advanced text editor to edit LilyPond sheet music files. " \
@@ -37,5 +37,7 @@ domain = "frescobaldi.org"
 url = "http://www.{0}/".format(domain)
 license = "GPL"
 
+required_python_ly_version = (0, 9, 4)
+
 # this one is used everywhere in the application
 appname = "Frescobaldi"
